<?php

if( !defined( 'ROOTPATH' ) )
    define( 'ROOTPATH', dirname(__FILE__).'/..' );

require_once(ROOTPATH.'/app/config.php');

/**
TODO list:

  * definir de forma centralizada os caminhos e as constantes necessárias;
  * criar um User Agent detect e um OS server detect para customizações espeçíficas de cada browser / servidor;
  * criar um registrador para fallback handlers;
  * criar um dependency manager na configuração dos serviços, para poder gerenciar os imports corretamente
  * criar um login e a recuperação da sessão;

*/

class Controller {

	static $cache;
	static $services = array();
	static $interceptors = array();
	static $config = array();
	static $includes = array();
	static $wallet;

	public function __destruct()
	{
// 	    if( $this->service )
// 		$this->service->close();
// 	    else
	    self::closeAll();
	}

	public static function closeAll()
	{
	    if( self::$services )
		foreach( self::$services as $serviceName => $service )
		    if( self::$config[ $serviceName ]['type'] === 'service' )
		      $service->close();
	}

	public static function clearAll()
	{
	    return self::$cache->clearAll();
	}

	public static function clear( $id )
	{
	    return self::$cache->clear( $id );
	}

	public static function get( $id )
	{
	    return self::$cache->get( $id );
	}

	public static function put( $id, $data, $expires, $compressed )
	{
	    return self::$cache->put( $id, $data, $expires, $compressed );
	}

	public static function find( $concept, $filter, $options = false )
	{
	    if( isset($URI['id']) && $URI['id'] )
		return self::read( $URI, $params, $options );

	    return self::call( 'find', $URI, $params, $options );
	}

	public static function read( $concept, $id, $options = false )
	{
	    if( !is_string( $id ) )
		return self::find( $concept, $id, $options );

	    return self::call( 'read', $concept, $id, $options );
	}
	
// 	public static function deleteAll( $URI, $params = false, $criteria = false )
// 	{
// 	    if( isset($URI['id']) && $URI['id'] )
// 		return self::delete( $URI, $params, $criteria );
// 
// 	    return self::call( 'deleteAll', $URI, $params, $criteria );
// 	}

	public static function delete( $concept, $id = false, $options = false )
	{
// 	    if( !isset($URI['id']) || !$URI['id'] )
// 		return self::deleteAll( $URI, $params, $criteria );

	    return self::call( 'delete', $URI, $params, $criteria );
	}
	
// 	public static function replace( $URI, $params, $criteria = false )
// 	{
// 	    if( isset($URI['id']) && $URI['id'] )
// 		return self::update( $URI, $params, $criteria );
// 
// 	    return self::call( 'replace', $URI, $params, $criteria );
// 	}

	public static function update( $concept, $params, $filter = false, $options = false )
	{
	    if( !isset($URI['id']) || !$URI['id'] )
		return self::replace( $URI, $params, $criteria );

	    return self::call( 'update', $URI, $params, $criteria );
	}

	public static function create( $concept, $params, $filter = false, $options = false )
	{
	    return self::call( 'create', $concept, $params, $filter, $options )
	}

	public static function begin( $concept, $id, $options = false )
	{
	    return self::call( 'begin', $concept, $id, $options );
	}

	public static function commit( $concept, /*$id,*/ $params = false, $options = false )
	{
	    return self::call( 'commit', $concept, $id, $params, $options );
	}

	public static function rollback($concept, /*$id,*/ $options = false )
	{
	    return self::call( 'rollback', $URI, $params, $options );
	}

	public static function format( $type, $data, $options = false )
	{
	    return self::call( 'format', $type, $data, $options );
	}

	public static function parse( $type, $data, $options = false )
	{
	    return self::call( 'parse', $type, $data, $options );
	}

	public static function URI( $className, $id = false, $service = false )
	{
	    return array( 'concept' => $className,
			  'service' => $service ? $service : false, 
			  'id' => $id ? $id : '' );
	}

	public static function links( $concept )
	{
	    if( !isset(self::$config[ $concept ]) )
	      self::$config[ $concept ] = self::loadConfig( $concept );

	    return( isset(self::$config[ $concept ]['links']) ? 
			  self::$config[ $concept ]['links'] : array() );
	}

	public static function isConcept( $concept )
	{
	    if( isset( self::$config[ $concept ] ) && 
		self::$config[ $concept ] )
		return( true );
	    else
		return file_exists( ROOTPATH."/config/$concept.ini" );
	}

	public static function getConcept( $concept )
	{
	    if( isset( self::$config[ $concept ] ) )
		return( self::$config[ $concept ] );

	    return( self::$config[ $concept ] = self::loadConfig( $concept ) );
	}

	public static function loadCache( $cacheType = 'Memory' )
	{
	    include_once( "cache/MemoryCache.php" );
	    return new MemoryCache();
	}
	
	public static function loadConfig( $className, $isService = false )
	{
	    $fileName = $className.'.'.($isService ? 'srv' : 'ini');

	    $config = self::$cache->get( $fileName );

	    if( !$config )
	    {
		$config = parse_ini_file( ROOTPATH."/config/$fileName", true );

		self::$cache->put( $fileName, $config );
	    }

	    return( $config );
	}

	public static function import( $path, $ext = ".php" )
	{
           
	    if( !isset(self::$includes[$path]) )
	    {
		require_once( ROOTPATH.'/'.$path.$ext );
		self::$includes[$path] = false;
	    }

	    return( self::$includes[$path] );
	}

	public static function load( $path, $class = false )
	{
            
            if( $return = self::import( $path, "" ) )
		return( $return );
            
	    if( !$class ){
		preg_match( '/^\/?.*\/([^\/]+).php$/', $path, $class );
		$class = $class[1];
	    }

	    $object =  self::$cache->get( $class );

	    if( !$object )
	    {
		$object = new $class();
		 self::$cache->put( $class, $object );
	    }

	    self::$includes[$path] = $object;

	    return( $object );
	}

	public static function wallet( $serviceName )
	{
	    if( !isset( self::$wallet ) )
	    {
		//// Hack //// TODO: passar o init da sessão no login do expresso
		if ( isset( $_COOKIE[ 'sessionid' ] ) ) 
		      session_id( $_COOKIE[ 'sessionid' ] ); 
                
		@session_start();

// 		if( !isset( $_SESSION['wallet'] ) )
// 		{
// 		    if( isset( $GLOBALS['phpgw_info'] ) )
// 		    {
// 				$_SESSION['wallet']['Sieve']['user']          = $GLOBALS['phpgw_info']['user']['account_lid'];
// 				$_SESSION['wallet']['Sieve']['password']      = $GLOBALS['phpgw_info']['user']['passwd'];
// 
// 				$_SESSION['wallet']['Cyrus']['user']          = $GLOBALS['phpgw_info']['user']['account_lid'];
// 				$_SESSION['wallet']['Cyrus']['password']      = $GLOBALS['phpgw_info']['user']['passwd'];
// 
// 				$_SESSION['wallet']['user']		      = $GLOBALS['phpgw_info']['user']['account_lid'];
// 			// 			$_SESSION['wallet']['me']['password']      = $_SESSION['phpgw_info']['expressomail']['user']['passwd'];
// 
// 				$_SESSION['wallet']['PostgreSQL']['user']     = $GLOBALS['phpgw_info']['server']['db_user'];
// 				$_SESSION['wallet']['PostgreSQL']['password'] = $GLOBALS['phpgw_info']['server']['db_pass'];
// 				$_SESSION['wallet']['PostgreSQL']['dbname']   = $GLOBALS['phpgw_info']['server']['db_name'];
// 		    }
// 		    else
// 			$_SESSION['wallet'] = array();
// 		}
                
                if(isset($_SESSION['wallet']))
                    self::$wallet = $_SESSION['wallet'];
		/////////////
	    }

	    return isset( self::$wallet[ $serviceName ] )? self::$wallet[ $serviceName ] : false;
	}

	public static function dispatch( $dispatcher, $data, $optionsMap = false )
	{
// 	    if( $mappedTo )
// 		$data = array( $mappedTo => $data );
// 
// 	    foreach( $data as $method => $params )
// 	    {
// // 		foreach( $data[ $method ] as $name => $value )
// 	    }
// 
// 	    self::import( "$dispatcher.php" );
	}

	public static function loadAll( $configurable, $isService = false )
	{
	}

	public static function connect( $service, $config )
	{
	    $result = $service->open( $config );

	    if( is_string( $result ) )
		throw new Exception( $result );

	    return( true );
	}

	public static function open( $serviceName, $config = false )
	{
	    $service = self::$services[ $serviceName ];

	    $srvConfig = array();

	    if( isset(self::$config[ $serviceName ][ 'config' ]) )
		$srvConfig = self::configure( $srvConfig, self::$config[ $serviceName ][ 'config' ] );
	    if( $wallet = self::wallet( $serviceName ) )
		$srvConfig = self::configure( $srvConfig, $wallet );
	    if( $config )
		$srvConfig = self::configure( $srvConfig, $config );

	    if( empty( $srvConfig ) )
		$srvConfig = false;

	    if( $service && self::$config[ $serviceName ]['type'] === 'service' )
		self::connect( $service, $srvConfig );
	}

	public static function configure( $config, $newConfig )
	{
	    foreach( $newConfig as $key => $value )
		$config[$key] = $value;

	    return( $config );
	}

	public static function service( $serviceName, $concept = false )
	{
	    if( isset( self::$services[ $serviceName ] ) )
		return self::$services[ $serviceName ];

	    if( !isset(self::$config[ $serviceName ]) )
		 self::$config[ $serviceName ] = self::loadConfig( $serviceName, true );

	    if( !isset(self::$config[ $serviceName ]) )
		return( false );

	    if( !isset(self::$config[ $serviceName ]['type']) )
		self::$config[ $serviceName ]['type'] = 'service';

	    self::import( 'app/'.self::$config[ $serviceName ]['type'] );   //TODO: Item 4

	    self::$services[ $serviceName ] = self::load( self::$config[ $serviceName ]['path'],
							  self::$config[ $serviceName ]['class'] );

	    if( $concept && isset(self::$config[ $concept ]['service.config']) )
		self::open( $serviceName, $self::$config[ $concept ]['service.config'] );
	    else
		self::open( $serviceName );

	    return( self::$services[ $serviceName ] = $service );
	}

	public static function interceptor( $method, $concept = false, $serviceName = false, $isService = false )
	{
	    if( $concept && !isset(self::$config[ $concept ]) )
	      self::$config[ $concept ] = self::loadConfig( $concept );

	    if( !$concept ) $concept = 'global';
	    if( !$isService || !$serviceName ) $serviceName = 'global';

	    if( !isset( self::$interceptors[ $concept ] ) )
		self::$interceptors[ $concept ] = array();

	    if( !isset( self::$interceptors[ $concept ][ $serviceName ] ) )
		self::$interceptors[ $concept ][ $serviceName ] = array();

	    if( !isset( self::$interceptors[ $concept ][ $serviceName ][ $method ] ) )
	    {
		$events = array( 'before', 'after' );
		$interceptors = array();

		$prefix = ( $isService )? "$serviceName." : "";

		foreach( $events as $i => $event )
		{
		    $interceptors[$event] = array();

		    if( !isset(self::$config[$concept]["$prefix$event.$method"]) )
		      continue;

		    foreach( self::$config[$concept]["$prefix$event.$method"] as $intercept => $interceptor )
			    $interceptors[$event][$intercept] = self::load( $interceptor );
		}

		self::$interceptors[ $concept ][ $serviceName ][ $method ] = $interceptors;
	    }

	    return( self::$interceptors[ $concept ][ $serviceName ][ $method ] );
	}

	public static function fire( $eventType, $method, $params, $original, $isService = false )
	{
	    $interceptors = self::interceptor( $method,
					       isset($original['URI']['concept']) ? $original['URI']['concept'] : false,
					       isset($params['URI']['service']) ? $params['URI']['service'] : false, $isService );

	    if( $interceptors && isset($interceptors[ $eventType ]) )
		foreach( $interceptors[ $eventType ] as $intercept => $interceptor )
		{
		    $return = $interceptor->$intercept( $params['URI'], $params['properties'], $params['criteria'], $original, $params['service'] );

		    if( $return === false )
			return( false );

		    if( isset($return) )
			$params['properties'] = $return;
		}

	      return( $params );
	}

	/*
	  * ex: array
	  *		(
	  *			[0] array( 'OR', array( array( '=', 'campo', 'valor' ), 
							  array( 'OR', array( array( '=', 'campo', 'valor' ) ) ) )
	  *			[1] array( '=', 'campo' , 'valor' )
	  *			[2] array( 'OR' , array( array( '=' , campo', 'valor' ) ) )
	  *			[3] array( 'IN', 'campo', array( '1' , '2' , '3' ) )
	  *		)
	  * OR
	  *	    array( '=' , 'campo' , 'valor' )
	*/

	public static function call( $method, $URI, $properties, $criteria = false, $service = false )
	{
	    try
	    {
		$original = $params = array( 'properties' => $properties, 
				             'criteria' => $criteria, 
					     'URI' => $URI,
					     'service' => $service );

		if( isset($params['URI']['concept']) )
		    $params = self::fire( 'before', $method, $params, $original );

		if( !$params['service'] )
		{
		    $isInternal = true;

		    if( ( !isset($params['URI']['service']) || !$params['URI']['service'] ) && isset($params['URI']['concept']) )
		    {
			if( !isset(self::$config[ $URI['concept'] ]) )
			    self::$config[ $URI['concept'] ] = self::loadConfig( $URI['concept'] );

			$params['URI']['service'] = self::$config[ $URI['concept'] ][ 'service' ];
		    }

		    if( !isset($params['URI']['service']) )
			return( false );

		    $params['service'] = self::service( $params['URI']['service'], isset($params['URI']['concept']) ? $params['URI']['concept'] : false );
		}

		if( isset($params['URI']['service']) )
		    $params = self::fire( 'before', $method, $params, $original, true );

		if( $params['service'] )
		    switch( $method )
		    {
			case 'find': $return = $params['service']->find( $params['URI'], $params['properties'], $params['criteria'] ); break;

			case 'read': $return = $params['service']->read( $params['URI'], $params['properties']/*, $criteria*/ ); break;

			case 'create': $return = $params['service']->create( $params['URI'], $params['properties']/*, $criteria*/ ); break;

			case 'delete': $return = $params['service']->delete( $params['URI'], $params['properties']/*, $criteria*/ ); break;

			case 'deleteAll': $return = $params['service']->deleteAll( $params['URI'], $params['properties'], $params['criteria'] ); break;

			case 'update': $return = $params['service']->update( $params['URI'], $params['properties']/*, $criteria*/ ); break;

			case 'replace': $return = $params['service']->replace( $params['URI'], $params['properties'], $params['criteria'] ); break;

			case 'begin': $return = $params['service']->begin( $params['URI'] ); break;

			case 'commit': $return = $params['service']->commit( $params['URI'] ); break;

			case 'rollback': $return = $params['service']->rollback( $params['URI'] ); break;

			case 'parse': $return = $params['service']->parse( $params['properties'], $params['criteria'] ); break;

			case 'format': $return = $params['service']->format( $params['properties'], $params['criteria'] ); break;
		    }
    
		if( isset($return) && $return !== false )
		    $params['properties'] = $return;

		if( isset($params['URI']['service']) )
		    $params = self::fire( 'after', $method, $params, $original, true );

		if( isset($URI['concept']) )
		    $params = self::fire( 'after', $method, $params, $original );
	    }
	    catch( Exception $e )
	    {
		if( !self::fallback( $e ) )
		    self::closeAll();

		return( false );
	    }

	    return( empty($params['properties']) ? false : $params['properties'] );
	}

	public static function fallback( $exception )
	{
	    error_log( $exception->getMessage() );
	    return( true );
	}
}

Controller::$cache = Controller::loadCache();
?>